
layout(early_fragment_tests) in;

out vec4 fcolor;

// Inputs from the VS
#ifndef INTERPOLATE
flat in vec3 WPos;
flat in vec3 WNormal;
flat in vec2 UV;
flat in vec3 ShadowPos;
#else
in vec3 WPos;
in vec3 WNormal;
in vec2 UV;
in vec3 ShadowPos;
#endif

// Uniforms
uniform vec3 CameraPos;
uniform vec3 ObjColor;
uniform sampler2D DiffTexture;
uniform sampler2D RoughnessTexture;
uniform samplerCube EnvMap;
uniform sampler2D CMaskTexture;
uniform vec3 SunDir;
uniform vec3 SunColor;

//uniform sampler2DShadow ShadowBuffer;
uniform sampler2D ShadowBuffer;

#define PI 3.14159f

// Using Hammersley sequence n
vec2 rand(vec2 px,float i,float n)
{	
	uint x = uint(px.x);
	uint y = uint(px.y);

	x = x * 1664525u + 1013904223u;
	y = y * 1664525u + 1013904223u;
	
	uint z = uint(1024*gl_FragCoord.z); 
	x += y*z;
	y += z*x;
	z += x*y;
	x += y*z;
	y += z*x;

	x = x >> 16u;
	y = y >> 16u;

	float E1 = fract( i / n + float( x & 0xffff ) / (1<<16) );
	float E2 = float( bitfieldReverse(uint(i)) ^ y ) * 2.3283064365386963e-10; // 1/2^32
	return vec2( E1, E2 );
}

float cheap_rand(float i)
{
	return fract(sin( dot(vec3(i,gl_FragCoord.xy), vec3(12.9898,78.233,45.164))) * 43758.5453);
}

// [Walter et al. 2007, "Microfacet models for refraction through rough surfaces"]
float D( float Alpha, float NoH )
{
	float a2 = Alpha * Alpha;
	float d = ( NoH * a2 - NoH ) * NoH + 1;
	return a2 / ( PI*d*d );					
}

// [Karis 2013, "Real Shading in Unreal Engine 4" ]
float G( float Alpha, float NoV, float NoL )
{
	float k = Alpha*0.5f;
	float Vis_SmithV = NoL / ( NoV * ( 1 - k ) + k );
	float Vis_SmithL = NoV / ( NoL * ( 1 - k ) + k );

	return Vis_SmithV*Vis_SmithL;
}

// [Schlick 1994, "An Inexpensive BRDF Model for Physically-Based Rendering"]
float F( float VoH )
{
	float iu = 1 - VoH;	
	float iu2 = iu*iu;
	float iu4 = iu2*iu2;
	iu = iu4*iu;

	const float fl = 0.5;
	return fl + (1.0 - fl)*iu;		
}

// [Karis 2013, "Real Shading in Unreal Engine 4" ]
vec3 ImportanceSampleGGX( vec2 Xi, float a, vec3 N )
{
	float Phi = 2 * PI * Xi.x;
	float CosTheta = sqrt( (1 - Xi.y) / ( 1 + (a*a - 1) * Xi.y ) );
	float SinTheta = sqrt( 1 - CosTheta * CosTheta );
	vec3 H;
	H.x = SinTheta * cos( Phi );
	H.y = SinTheta * sin( Phi );
	H.z = CosTheta;
	vec3 UpVector = abs(N.z) < 0.999 ? vec3(0,0,1) : vec3(1,0,0);
	vec3 TangentX = normalize( cross( UpVector, N ) );
	vec3 TangentY = cross( N, TangentX );
	// Tangent to world space
	return TangentX * H.x + TangentY * H.y + N * H.z;
}

// It's much more efficient to split this and pre-compute, but I didn't feel like learning how you can do compute shaders on OpenGL
// Not enough time, and with 8 samples it works ok
// I'm crying inside, specially because this is done on a forward-shading pipeline, but not enough time to implement everything
// [Karis 2013, "Real Shading in Unreal Engine 4" ]
vec3 SpecularIBL( float alpha, vec3 N, vec3 V )
{
	vec3 SpecularLighting = vec3(0);
	const float NumSamples = 4;
	for( float i = 0; i < NumSamples; i++ )
	{
		vec3 H = ImportanceSampleGGX( rand(gl_FragCoord.xy,i,NumSamples), alpha, N );
		
		float lerp_r = 0.3f;
		H = H*lerp_r + (1.0f - lerp_r)*N;

		vec3 L = 2 * dot( V, H ) * H - V;
		float NoV = clamp( dot( N, V ) ,0,1);
		float NoL = clamp( dot( N, L ) ,0,1);
		float NoH = clamp( dot( N, H ) ,0,1);
		float VoH = clamp( dot( V, H ) ,0,1);
		if( NoL > 0 )
		{
			vec3 SampleColor = textureLod(EnvMap,L,10*sqrt(alpha)).rgb;

			// Incident light = SampleColor * NoL
			// Microfacet specular = D*G*F / (4*NoL*NoV)
			// pdf = D * NoH / (4 * VoH)
			SpecularLighting += SampleColor  * G(alpha,NoV,NoL) * F(VoH) * VoH / (NoH * NoV);
		}
	}
	return SpecularLighting * (1/ NumSamples);
}

float lerp(float x,float a,float b)
{
	return x*(1-a) + b*x;
}


float SampleShadow(float BaseBias,float NoL, const float SamplesN,const float Radius)
{
	float ret = 0;
	NoL = clamp(NoL,0.01f,1);
	float a = 0.25f;
	float bias = BaseBias*((1 - NoL)/(a*NoL) + 1); // Approximates tan(acos(NoL))

	for(float i = 0;i < SamplesN;i++)
	{
		float z0 = ShadowPos.z;
		float zs = texture( ShadowBuffer, ShadowPos.xy + Radius*rand(gl_FragCoord.xy,i,SamplesN)).r;
		
		// Based on the integral of zs <= z0 - b*r ? 0.0f : 1.0f, with respect to r in the [0.1,1] range
		// Ajusted with a power to make it softer
		ret += pow(clamp((bias - z0 + zs)/bias,0.0f,1.0f),0.7);
	}

	return clamp(ret * (1.0f / SamplesN),0.0f,1.0f);
}

void main()
{
	fcolor = vec4(0);

	float roughness = clamp(texture(RoughnessTexture,UV).r,0.05,1.0f);
	float alpha = roughness*roughness;

	vec3 n = normalize(WNormal);
	vec3 v = normalize(CameraPos - WPos);
	vec3 r = reflect(v,WNormal);
	vec3 h = normalize(v + SunDir);
	float NoV = clamp( dot( n, v ) ,0,1);
	float NoL = clamp( dot( n, SunDir ) ,0,1);
	float NoH = clamp( dot( n, h ) ,0,1);
	float VoH = clamp( dot( v, h ) ,0,1);

	// Not exactly albedo, as it has AO baked in...
	vec3 albedo = texture(DiffTexture,UV).rgb;
	albedo = clamp(albedo,0.1f,1.0f);

	float cmask = texture(CMaskTexture,UV).r;
	albedo *= (1.0f - cmask) + cmask*ObjColor;

	vec3 refl = SpecularIBL(alpha,n,v);
	vec3 ambient = textureLod(EnvMap,n,9).rgb; // Using a coarse mip as a preintegrated ambient

	vec3 diff = SunColor*clamp(dot(n,SunDir),0.0f,1.0f);

	// And this right here is what totally breaks all the nice PBR ideology, but what you gonna do :/
	float ka = 1.5f;
	float kr = 0.45f;
	float kd = 1.0f;
	float ks = 0.3f;
	float kl = 1.5f;

	fcolor.rgb = ka*ambient*albedo + kr*refl + SampleShadow(0.0003f,NoL,8,0.0004f)*kl*SunColor*NoL*(kd*albedo + ks*vec3(D(alpha,NoH)*G(alpha,NoV,NoL)*F(VoH)));
	fcolor.a = 1.0f;
}